@@ -6,6 +6,7 @@
  */
 
 #include "gulkan-device.h"
+#include "gulkan-queue.h"
 
 struct _GulkanDevice
 {
@@ -14,12 +15,11 @@ struct _GulkanDevice
   VkDevice device;
   VkPhysicalDevice physical_device;
 
-  VkQueue queue;
-
-  uint32_t queue_family_index;
-
   VkPhysicalDeviceMemoryProperties memory_properties;
 
+  GulkanQueue *graphics_queue;
+  GulkanQueue *transfer_queue;
+
   PFN_vkGetMemoryFdKHR extVkGetMemoryFdKHR;
 };
 
@@ -30,7 +30,8 @@ gulkan_device_init (GulkanDevice *self)
 {
   self->device = VK_NULL_HANDLE;
   self->physical_device = VK_NULL_HANDLE;
-  self->queue = VK_NULL_HANDLE;
+  self->transfer_queue = NULL;
+  self->graphics_queue = NULL;
   self->extVkGetMemoryFdKHR = 0;
 }
 
@@ -44,7 +45,11 @@ static void
 gulkan_device_finalize (GObject *gobject)
 {
   GulkanDevice *self = GULKAN_DEVICE (gobject);
+  if (self->graphics_queue) g_object_unref (self->graphics_queue);
+  if (self->transfer_queue) g_object_unref (self->transfer_queue);
   vkDestroyDevice (self->device, NULL);
+
+  G_OBJECT_CLASS (gulkan_device_parent_class)->finalize (gobject);
 }
 
 static void
@@ -55,7 +60,7 @@ gulkan_device_class_init (GulkanDeviceClass *klass)
   object_class->finalize = gulkan_device_finalize;
 }
 
-static bool
+static gboolean
 _find_physical_device (GulkanDevice    *self,
                        GulkanInstance  *instance,
                        VkPhysicalDevice requested_device)
@@ -64,12 +69,12 @@ _find_physical_device (GulkanDevice    *self,
   uint32_t num_devices = 0;
   VkResult res =
     vkEnumeratePhysicalDevices (vk_instance, &num_devices, NULL);
-  vk_check_error ("vkEnumeratePhysicalDevices", res, false)
+  vk_check_error ("vkEnumeratePhysicalDevices", res, FALSE);
 
   if (num_devices == 0)
     {
       g_printerr ("No Vulkan devices found.\n");
-      return false;
+      return FALSE;
     }
 
   VkPhysicalDevice *physical_devices =
@@ -77,7 +82,7 @@ _find_physical_device (GulkanDevice    *self,
   res = vkEnumeratePhysicalDevices (vk_instance,
                                    &num_devices,
                                     physical_devices);
-  vk_check_error ("vkEnumeratePhysicalDevices", res, false)
+  vk_check_error ("vkEnumeratePhysicalDevices", res, FALSE);
 
   if (requested_device == VK_NULL_HANDLE)
     {
@@ -108,11 +113,11 @@ _find_physical_device (GulkanDevice    *self,
   vkGetPhysicalDeviceMemoryProperties (self->physical_device,
                                       &self->memory_properties);
 
-  return true;
+  return TRUE;
 }
 
-static bool
-_find_graphics_queue (GulkanDevice *self)
+static gboolean
+_find_queue_families (GulkanDevice *self)
 {
   /* Find the first graphics queue */
   uint32_t num_queues = 0;
@@ -128,7 +133,7 @@ _find_graphics_queue (GulkanDevice *self)
   if (num_queues == 0)
     {
       g_printerr ("Failed to get queue properties.\n");
-      return false;
+      return FALSE;
     }
 
   uint32_t i = 0;
@@ -139,38 +144,42 @@ _find_graphics_queue (GulkanDevice *self)
   if (i >= num_queues)
     {
       g_printerr ("No graphics queue found\n");
-      return false;
+      return FALSE;
     }
-  self->queue_family_index = i;
+ uint32_t graphics_queue_index = i;
+
+  for (i = 0; i < num_queues; i++)
+    if (i != graphics_queue_index)
+      if (queue_family_props[i].queueFlags & VK_QUEUE_TRANSFER_BIT)
+          break;
+
+  if (i >= num_queues)
+    {
+      g_printerr ("No transfer queue found\n");
+      return FALSE;
+    }
+  uint32_t transfer_queue_index = i;
 
   g_free (queue_family_props);
 
-  return true;
+  self->graphics_queue = gulkan_queue_new (self, graphics_queue_index);
+  self->transfer_queue = gulkan_queue_new (self, transfer_queue_index);
+
+  return TRUE;
 }
 
-bool
-gulkan_device_queue_supports_surface (GulkanDevice *self,
-                                      VkSurfaceKHR  surface)
-{
-  VkBool32 surface_support = false;
-  vkGetPhysicalDeviceSurfaceSupportKHR (self->physical_device,
-                                        self->queue_family_index,
-                                        surface, &surface_support);
-  return surface_support;
-}
-
-static bool
+static gboolean
 _get_device_extension_count (GulkanDevice *self,
                              uint32_t     *count)
 {
   VkResult res =
     vkEnumerateDeviceExtensionProperties (self->physical_device, NULL,
                                           count, NULL);
-  vk_check_error ("vkEnumerateDeviceExtensionProperties", res, false)
-  return true;
+  vk_check_error ("vkEnumerateDeviceExtensionProperties", res, FALSE);
+  return TRUE;
 }
 
-static bool
+static gboolean
 _init_device_extensions (GulkanDevice *self,
                          GSList       *required_extensions,
                          uint32_t      num_extensions,
@@ -189,17 +198,17 @@ _init_device_extensions (GulkanDevice *self,
                                                        NULL,
                                                       &num_extensions,
                                                        extension_props);
-  vk_check_error ("vkEnumerateDeviceExtensionProperties", res, false)
+  vk_check_error ("vkEnumerateDeviceExtensionProperties", res, FALSE);
   for (uint32_t i = 0; i < g_slist_length (required_extensions); i++)
     {
-      bool found = false;
+      gboolean found = FALSE;
       for (uint32_t j = 0; j < num_extensions; j++)
         {
           GSList* extension_name = g_slist_nth (required_extensions, i);
           if (strcmp ((gchar*) extension_name->data,
                       extension_props[j].extensionName) == 0)
             {
-              found = true;
+              found = TRUE;
               break;
             }
         }
@@ -217,24 +226,24 @@ _init_device_extensions (GulkanDevice *self,
 
   g_free (extension_props);
 
-  return true;
+  return TRUE;
 }
 
-bool
+gboolean
 gulkan_device_create (GulkanDevice    *self,
                       GulkanInstance  *instance,
                       VkPhysicalDevice device,
                       GSList          *extensions)
 {
   if (!_find_physical_device (self, instance, device))
-    return false;
+    return FALSE;
 
-  if (!_find_graphics_queue (self))
-    return false;
+  if (!_find_queue_families (self))
+    return FALSE;
 
   uint32_t num_extensions = 0;
   if (!_get_device_extension_count (self, &num_extensions))
-    return false;
+    return FALSE;
 
   char **extension_names = g_malloc (sizeof (char*) * num_extensions);
   uint32_t num_enabled = 0;
@@ -243,18 +252,16 @@ gulkan_device_create (GulkanDevice    *self,
     {
       if (!_init_device_extensions (self, extensions, num_extensions,
                                     extension_names, &num_enabled))
-        return false;
+        return FALSE;
     }
 
   if (num_enabled > 0)
     {
-      g_print ("Requesting device extensions:\n");
+      g_debug ("Requesting device extensions:");
       for (uint32_t i = 0; i < num_enabled; i++)
-          g_print ("%s\n", extension_names[i]);
+        g_debug ("%s", extension_names[i]);
     }
 
-  float queue_priority = 1.0f;
-
   VkPhysicalDeviceFeatures physical_device_features;
   vkGetPhysicalDeviceFeatures (self->physical_device,
                               &physical_device_features);
@@ -262,14 +269,21 @@ gulkan_device_create (GulkanDevice    *self,
   VkDeviceCreateInfo device_info =
     {
       .sType = VK_STRUCTURE_TYPE_DEVICE_CREATE_INFO,
-      .queueCreateInfoCount = 1,
-      .pQueueCreateInfos = &(VkDeviceQueueCreateInfo)
+      .queueCreateInfoCount = 2,
+      .pQueueCreateInfos = (VkDeviceQueueCreateInfo[]) {
         {
           .sType = VK_STRUCTURE_TYPE_DEVICE_QUEUE_CREATE_INFO,
-          .queueFamilyIndex = self->queue_family_index,
+          .queueFamilyIndex = gulkan_queue_get_family_index(self->graphics_queue),
           .queueCount = 1,
-          .pQueuePriorities = &queue_priority
+          .pQueuePriorities = (const float[]) { 1.0f }
         },
+        {
+          .sType = VK_STRUCTURE_TYPE_DEVICE_QUEUE_CREATE_INFO,
+          .queueFamilyIndex = gulkan_queue_get_family_index (self->transfer_queue),
+          .queueCount = 1,
+          .pQueuePriorities = (const float[]) { 0.8f }
+        },
+      },
       .enabledExtensionCount = num_enabled,
       .ppEnabledExtensionNames = (const char* const*) extension_names,
       .pEnabledFeatures = &physical_device_features
@@ -277,9 +291,13 @@ gulkan_device_create (GulkanDevice    *self,
 
   VkResult res = vkCreateDevice (self->physical_device,
                                 &device_info, NULL, &self->device);
-  vk_check_error ("vkCreateDevice", res, false)
+  vk_check_error ("vkCreateDevice", res, FALSE);
 
-  vkGetDeviceQueue (self->device, self->queue_family_index, 0, &self->queue);
+  if (!gulkan_queue_initialize (self->graphics_queue))
+      return FALSE;
+
+  if (!gulkan_queue_initialize (self->transfer_queue))
+    return FALSE;
 
   if (num_enabled > 0)
     {
@@ -288,10 +306,10 @@ gulkan_device_create (GulkanDevice    *self,
     }
   g_free (extension_names);
 
-  return true;
+  return TRUE;
 }
 
-bool
+gboolean
 gulkan_device_memory_type_from_properties (
   GulkanDevice         *self,
   uint32_t              memory_type_bits,
@@ -306,107 +324,14 @@ gulkan_device_memory_type_from_properties (
            & memory_property_flags) == memory_property_flags)
       {
         *type_index_out = i;
-        return true;
+        return TRUE;
       }
     }
     memory_type_bits >>= 1;
   }
 
   /* Could not find matching memory type.*/
-  return false;
-}
-
-bool
-gulkan_device_create_buffer (GulkanDevice         *self,
-                             VkDeviceSize          size,
-                             VkBufferUsageFlags    usage,
-                             VkMemoryPropertyFlags properties,
-                             VkBuffer             *buffer,
-                             VkDeviceMemory       *memory)
-{
-  VkBufferCreateInfo buffer_info =
-  {
-    .sType = VK_STRUCTURE_TYPE_BUFFER_CREATE_INFO,
-    .size = size,
-    .usage = usage,
-  };
-  VkResult res = vkCreateBuffer (self->device, &buffer_info, NULL, buffer);
-  vk_check_error ("vkCreateBuffer", res, false)
-
-  VkMemoryRequirements requirements = {};
-  vkGetBufferMemoryRequirements (self->device, *buffer, &requirements);
-
-  VkMemoryAllocateInfo alloc_info = {
-    .sType = VK_STRUCTURE_TYPE_MEMORY_ALLOCATE_INFO,
-    .allocationSize = requirements.size
-  };
-
-  if (!gulkan_device_memory_type_from_properties (
-        self,
-        requirements.memoryTypeBits,
-        properties,
-        &alloc_info.memoryTypeIndex))
-  {
-    g_printerr ("Failed to find matching memoryTypeIndex for buffer\n");
-    return false;
-  }
-
-  res = vkAllocateMemory (self->device, &alloc_info, NULL, memory);
-  vk_check_error ("vkAllocateMemory", res, false)
-
-  res = vkBindBufferMemory (self->device, *buffer, *memory, 0);
-  vk_check_error ("vkBindBufferMemory", res, false)
-
-  return true;
-}
-
-bool
-gulkan_device_create_buffer_from_data (GulkanDevice         *self,
-                                       const void           *data,
-                                       VkDeviceSize          size,
-                                       VkBufferUsageFlags    usage,
-                                       VkMemoryPropertyFlags properties,
-                                       VkBuffer             *buffer,
-                                       VkDeviceMemory       *memory)
-{
-  if (!gulkan_device_create_buffer (self, size, usage,
-                                    properties, buffer, memory))
-    return false;
-
-  if (!gulkan_device_map_memory (self, data, size, *memory))
-    return false;
-
-  return true;
-}
-
-bool
-gulkan_device_map_memory (GulkanDevice  *self,
-                          const void    *data,
-                          VkDeviceSize   size,
-                          VkDeviceMemory memory)
-{
-  if (data == NULL)
-    {
-      g_printerr ("Trying to map NULL memory.\n");
-      return false;
-    }
-
-  void *tmp;
-  VkResult res = vkMapMemory (self->device, memory, 0, VK_WHOLE_SIZE, 0, &tmp);
-  vk_check_error ("vkMapMemory", res, false)
-
-  memcpy (tmp, data, size);
-  vkUnmapMemory (self->device, memory);
-
-  VkMappedMemoryRange memory_range = {
-    .sType = VK_STRUCTURE_TYPE_MAPPED_MEMORY_RANGE,
-    .memory = memory,
-    .size = VK_WHOLE_SIZE
-  };
-  res = vkFlushMappedMemoryRanges (self->device, 1, &memory_range);
-  vk_check_error ("vkFlushMappedMemoryRanges", res, false)
-
-  return true;
+  return FALSE;
 }
 
 VkDevice
@@ -421,16 +346,22 @@ gulkan_device_get_physical_handle (GulkanDevice *self)
   return self->physical_device;
 }
 
-uint32_t
-gulkan_device_get_queue_family_index (GulkanDevice *self)
+GulkanQueue*
+gulkan_device_get_graphics_queue (GulkanDevice *self)
 {
-  return self->queue_family_index;
+  return self->graphics_queue;
+}
+
+GulkanQueue*
+gulkan_device_get_transfer_queue (GulkanDevice *self)
+{
+  return self->transfer_queue;
 }
 
 gboolean
-gulkan_device_gulkan_device_get_memory_fd (GulkanDevice *self,
-                                           VkDeviceMemory image_memory,
-                                           int *fd)
+gulkan_device_get_memory_fd (GulkanDevice  *self,
+                             VkDeviceMemory image_memory,
+                             int           *fd)
 {
   if (!self->extVkGetMemoryFdKHR)
     self->extVkGetMemoryFdKHR =
@@ -459,12 +390,6 @@ gulkan_device_gulkan_device_get_memory_fd (GulkanDevice *self,
   return TRUE;
 }
 
-VkQueue
-gulkan_device_get_queue_handle (GulkanDevice *self)
-{
-  return self->queue;
-}
-
 void
 gulkan_device_wait_idle (GulkanDevice *self)
 {
