@@ -5,8 +5,8 @@
  * SPDX-License-Identifier: MIT
  */
 
-#ifndef XRD_GEOMETRY_H_
-#define XRD_GEOMETRY_H_
+#ifndef GULKAN_GEOMETRY_H_
+#define GULKAN_GEOMETRY_H_
 
 #if !defined (GULKAN_INSIDE) && !defined (GULKAN_COMPILATION)
 #error "Only <gulkan.h> can be included directly."
@@ -34,4 +34,4 @@ gulkan_geometry_append_plane (GulkanVertexBuffer *self,
                               graphene_point_t   *to,
                               graphene_matrix_t  *mat);
 
-#endif /* XRD_GEOMETRY_H_ */
+#endif /* GULKAN_GEOMETRY_H_ */
