@@ -14,23 +14,24 @@
 
 #include <glib-object.h>
 #include <vulkan/vulkan.h>
-#include <stdbool.h>
 
 G_BEGIN_DECLS
 
 const gchar*
 vk_result_string (VkResult code);
 
+const gchar*
+vk_format_string (VkFormat format);
+
+gboolean
+gulkan_has_error (VkResult     res,
+                  const gchar *fun,
+                  const gchar *file,
+                  int          line);
+
 #define vk_check_error(fun, res, ret) \
-{ \
-  VkResult r = (res); \
-  if (r != VK_SUCCESS) \
-  { \
-    g_printerr ("ERROR: " fun " failed with %s in %s:%d\n", \
-                vk_result_string (r), __FILE__, __LINE__); \
-    return (ret); \
-  }\
-}
+if (gulkan_has_error (res, fun, __FILE__, __LINE__)) \
+  return ret
 
 #define GULKAN_TYPE_INSTANCE gulkan_instance_get_type ()
 G_DECLARE_FINAL_TYPE (GulkanInstance, gulkan_instance,
@@ -38,7 +39,7 @@ G_DECLARE_FINAL_TYPE (GulkanInstance, gulkan_instance,
 
 GulkanInstance *gulkan_instance_new (void);
 
-bool
+gboolean
 gulkan_instance_create (GulkanInstance *self,
                         GSList* requested_extensions);
 
