@@ -9,26 +9,24 @@
 #include "gulkan-texture.h"
 
 #include <vulkan/vulkan.h>
-#include <stdbool.h>
+#include "gulkan-buffer.h"
+#include "gulkan-cmd-buffer.h"
 
 struct _GulkanTexture
 {
-  GObjectClass parent_class;
+  GObjectClass parent;
+
+  GulkanClient *client;
 
   VkImage image;
   VkDeviceMemory image_memory;
   VkImageView image_view;
-  VkBuffer staging_buffer;
-  VkDeviceMemory staging_buffer_memory;
 
   guint mip_levels;
 
-  uint32_t width;
-  uint32_t height;
+  VkExtent2D extent;
 
   VkFormat format;
-
-  GulkanDevice *device;
 };
 
 G_DEFINE_TYPE (GulkanTexture, gulkan_texture, G_TYPE_OBJECT)
@@ -43,65 +41,107 @@ typedef struct {
 static GulkanMipMap
 _generate_mipmaps (GdkPixbuf *pixbuf);
 
-static bool
-_upload_pixels (GulkanTexture           *self,
-                VkCommandBuffer          cmd_buffer,
-                guchar                  *pixels,
-                gsize                    size,
-                const VkBufferImageCopy *regions,
-                VkImageLayout            layout);
-
 static void
 gulkan_texture_init (GulkanTexture *self)
 {
   self->image = VK_NULL_HANDLE;
   self->image_memory = VK_NULL_HANDLE;
   self->image_view = VK_NULL_HANDLE;
-  self->staging_buffer = VK_NULL_HANDLE;
-  self->staging_buffer_memory = VK_NULL_HANDLE;
   self->format = VK_NULL_HANDLE;
   self->mip_levels = 1;
 }
 
 static void
-gulkan_texture_finalize (GObject *gobject)
+_finalize (GObject *gobject)
 {
   GulkanTexture *self = GULKAN_TEXTURE (gobject);
-  VkDevice device = gulkan_device_get_handle (self->device);
-
-  gulkan_texture_free_staging_memory (self);
+  VkDevice device = gulkan_client_get_device_handle (self->client);
 
   vkDestroyImageView (device, self->image_view, NULL);
   vkDestroyImage (device, self->image, NULL);
   vkFreeMemory (device, self->image_memory, NULL);
-}
 
-void
-gulkan_texture_free_staging_memory (GulkanTexture *self)
-{
-  VkDevice device = gulkan_device_get_handle (self->device);
-  vkDestroyBuffer (device, self->staging_buffer, NULL);
-  vkFreeMemory (device, self->staging_buffer_memory, NULL);
+  g_object_unref (self->client);
+
+  G_OBJECT_CLASS (gulkan_texture_parent_class)->finalize (gobject);
 }
 
 static void
 gulkan_texture_class_init (GulkanTextureClass *klass)
 {
   GObjectClass *object_class = G_OBJECT_CLASS (klass);
+  object_class->finalize = _finalize;
+}
 
-  object_class->finalize = gulkan_texture_finalize;
+static gboolean
+_upload_pixels (GulkanTexture           *self,
+                guchar                  *pixels,
+                gsize                    size,
+                const VkBufferImageCopy *regions,
+                VkImageLayout            layout)
+{
+  GulkanDevice *device = gulkan_client_get_device (self->client);
+  GulkanQueue *queue = gulkan_device_get_transfer_queue (device);
+  GulkanCmdBuffer *cmd_buffer = gulkan_queue_request_cmd_buffer (queue);
+  GMutex *mutex = gulkan_queue_get_pool_mutex (queue);
+
+  g_mutex_lock (mutex);
+  gboolean ret = gulkan_cmd_buffer_begin (cmd_buffer);
+  g_mutex_unlock (mutex);
+
+  if (!ret)
+    return FALSE;
+
+  GulkanBuffer *staging_buffer =
+    gulkan_buffer_new (device, size,
+                       VK_BUFFER_USAGE_TRANSFER_SRC_BIT,
+                       VK_MEMORY_PROPERTY_HOST_VISIBLE_BIT);
+
+  if (!staging_buffer)
+    return FALSE;
+
+  if (!gulkan_buffer_upload (staging_buffer, pixels, size))
+    return FALSE;
+
+  gulkan_texture_record_transfer (self,
+                                  gulkan_cmd_buffer_get_handle (cmd_buffer),
+                                  VK_IMAGE_LAYOUT_UNDEFINED,
+                                  VK_IMAGE_LAYOUT_TRANSFER_DST_OPTIMAL);
+
+  g_mutex_lock (mutex);
+  vkCmdCopyBufferToImage (gulkan_cmd_buffer_get_handle (cmd_buffer),
+                          gulkan_buffer_get_handle (staging_buffer),
+                          self->image,
+                          VK_IMAGE_LAYOUT_TRANSFER_DST_OPTIMAL,
+                          self->mip_levels,
+                          regions);
+  g_mutex_unlock (mutex);
+
+  gulkan_texture_record_transfer (self,
+                                  gulkan_cmd_buffer_get_handle (cmd_buffer),
+                                  VK_IMAGE_LAYOUT_TRANSFER_DST_OPTIMAL,
+                                  layout);
+
+  if (!gulkan_queue_submit (queue, cmd_buffer))
+    return FALSE;
+
+  g_object_unref (staging_buffer);
+  gulkan_queue_free_cmd_buffer (queue, cmd_buffer);
+
+  return TRUE;
 }
 
 GulkanTexture *
-gulkan_texture_new_from_pixbuf (GulkanDevice   *device,
-                                VkCommandBuffer cmd_buffer,
+gulkan_texture_new_from_pixbuf (GulkanClient   *client,
                                 GdkPixbuf      *pixbuf,
                                 VkFormat        format,
                                 VkImageLayout   layout,
-                                bool            create_mipmaps)
+                                gboolean        create_mipmaps)
 {
-  guint width = (guint) gdk_pixbuf_get_width (pixbuf);
-  guint height = (guint) gdk_pixbuf_get_height (pixbuf);
+  VkExtent2D extent = {
+    .width = (uint32_t) gdk_pixbuf_get_width (pixbuf),
+    .height = (uint32_t) gdk_pixbuf_get_height (pixbuf)
+  };
 
   GulkanTexture *self;
 
@@ -109,11 +149,11 @@ gulkan_texture_new_from_pixbuf (GulkanDevice   *device,
     {
       GulkanMipMap mipmap = _generate_mipmaps (pixbuf);
 
-      self = gulkan_texture_new_mip_levels (device, width, height,
-                                            mipmap.levels, mipmap.size, format);
+      self = gulkan_texture_new_mip_levels (client, extent,
+                                            mipmap.levels, format);
 
-      if(!_upload_pixels (self, cmd_buffer, mipmap.buffer,
-                          mipmap.size, mipmap.buffer_image_copies, layout))
+      if(!_upload_pixels (self, mipmap.buffer, mipmap.size,
+                          mipmap.buffer_image_copies, layout))
         {
           g_printerr ("ERROR: Could not upload pixels.\n");
           g_object_unref (self);
@@ -127,10 +167,9 @@ gulkan_texture_new_from_pixbuf (GulkanDevice   *device,
     {
       gsize size = gdk_pixbuf_get_byte_length (pixbuf);
       guchar *pixels = gdk_pixbuf_get_pixels (pixbuf);
-      self = gulkan_texture_new (device, width, height, size, format);
+      self = gulkan_texture_new (client, extent, format);
 
-      if (!gulkan_texture_upload_pixels (self, cmd_buffer,
-                                         pixels, size, layout))
+      if (!gulkan_texture_upload_pixels (self, pixels, size, layout))
         {
           g_printerr ("ERROR: Could not upload pixels.\n");
           g_object_unref (self);
@@ -142,25 +181,25 @@ gulkan_texture_new_from_pixbuf (GulkanDevice   *device,
 }
 
 GulkanTexture *
-gulkan_texture_new_from_cairo_surface (GulkanDevice    *device,
-                                       VkCommandBuffer  cmd_buffer,
+gulkan_texture_new_from_cairo_surface (GulkanClient    *client,
                                        cairo_surface_t *surface,
                                        VkFormat         format,
                                        VkImageLayout    layout)
 {
-  guint width = (guint)cairo_image_surface_get_width (surface);
-  guint height = (guint)cairo_image_surface_get_height (surface);
+  VkExtent2D extent = {
+    .width = (uint32_t) cairo_image_surface_get_width (surface),
+    .height = (uint32_t) cairo_image_surface_get_height (surface)
+  };
 
   guint stride = (guint)cairo_image_surface_get_stride (surface);
 
-  guint size = stride * height;
+  guint size = stride * extent.height;
 
   guchar *pixels = cairo_image_surface_get_data (surface);
 
-  GulkanTexture *self =
-    gulkan_texture_new (device, width, height, size, format);
+  GulkanTexture *self = gulkan_texture_new (client, extent, format);
 
-  if (!gulkan_texture_upload_pixels (self, cmd_buffer, pixels, size, layout))
+  if (!gulkan_texture_upload_pixels (self, pixels, size, layout))
     {
       g_printerr ("ERROR: Could not upload pixels.\n");
       g_object_unref (self);
@@ -171,53 +210,51 @@ gulkan_texture_new_from_cairo_surface (GulkanDevice    *device,
 }
 
 GulkanTexture *
-gulkan_texture_new (GulkanDevice *device,
-                    guint         width,
-                    guint         height,
-                    gsize         size,
+gulkan_texture_new (GulkanClient *client,
+                    VkExtent2D    extent,
                     VkFormat      format)
 {
-  return gulkan_texture_new_mip_levels (device, width, height, 1, size, format);
+  return gulkan_texture_new_mip_levels (client, extent, 1, format);
 }
 
 GulkanTexture *
-gulkan_texture_new_mip_levels (GulkanDevice *device,
-                               guint         width,
-                               guint         height,
+gulkan_texture_new_mip_levels (GulkanClient *client,
+                               VkExtent2D    extent,
                                guint         mip_levels,
-                               gsize         size,
                                VkFormat      format)
 {
   GulkanTexture *self = (GulkanTexture*) g_object_new (GULKAN_TYPE_TEXTURE, 0);
 
-  self->width = width;
-  self->height = height;
-  self->device = device;
+  self->extent = extent;
+  self->client = g_object_ref (client);
   self->format = format;
   self->mip_levels = mip_levels;
 
-  VkDevice vk_device = gulkan_device_get_handle (device);
+  VkDevice vk_device = gulkan_client_get_device_handle (client);
 
   /* TODO: Check with vkGetPhysicalDeviceFormatProperties */
   VkImageTiling tiling;
   switch (format)
     {
+    case VK_FORMAT_R8G8B8_SRGB:
     case VK_FORMAT_R8G8B8_UNORM:
       tiling = VK_IMAGE_TILING_LINEAR;
       break;
-    case  VK_FORMAT_R8G8B8A8_UNORM:
+    case VK_FORMAT_R8G8B8A8_SRGB:
+    case VK_FORMAT_R8G8B8A8_UNORM:
       tiling = VK_IMAGE_TILING_OPTIMAL;
       break;
     default:
-      g_printerr ("Error: No tiling option for format %d\n", format);
+      g_printerr ("Warning: No tiling for format %s (%d) specified.\n",
+                  vk_format_string(format), format);
       tiling = VK_IMAGE_TILING_OPTIMAL;
     }
 
   VkImageCreateInfo image_info = {
     .sType = VK_STRUCTURE_TYPE_IMAGE_CREATE_INFO,
     .imageType = VK_IMAGE_TYPE_2D,
-    .extent.width = width,
-    .extent.height = height,
+    .extent.width = extent.width,
+    .extent.height = extent.height,
     .extent.depth = 1,
     .mipLevels = mip_levels,
     .arrayLayers = 1,
@@ -231,9 +268,9 @@ gulkan_texture_new_mip_levels (GulkanDevice *device,
   };
   VkResult res;
   res = vkCreateImage (vk_device, &image_info, NULL, &self->image);
-  vk_check_error ("vkCreateImage", res, NULL)
+  vk_check_error ("vkCreateImage", res, NULL);
 
-  VkMemoryRequirements memory_requirements = {};
+  VkMemoryRequirements memory_requirements;
   vkGetImageMemoryRequirements (vk_device, self->image,
                                 &memory_requirements);
 
@@ -243,15 +280,15 @@ gulkan_texture_new_mip_levels (GulkanDevice *device,
     .allocationSize = memory_requirements.size
   };
   gulkan_device_memory_type_from_properties (
-    device,
+    gulkan_client_get_device (client),
     memory_requirements.memoryTypeBits,
     VK_MEMORY_PROPERTY_DEVICE_LOCAL_BIT,
     &memory_info.memoryTypeIndex);
   res = vkAllocateMemory (vk_device, &memory_info,
                           NULL, &self->image_memory);
-  vk_check_error ("vkAllocateMemory", res, NULL)
+  vk_check_error ("vkAllocateMemory", res, NULL);
   res = vkBindImageMemory (vk_device, self->image, self->image_memory, 0);
-  vk_check_error ("vkBindImageMemory", res, NULL)
+  vk_check_error ("vkBindImageMemory", res, NULL);
 
   VkImageViewCreateInfo image_view_info =
   {
@@ -269,47 +306,11 @@ gulkan_texture_new_mip_levels (GulkanDevice *device,
   };
   res = vkCreateImageView (vk_device, &image_view_info,
                            NULL, &self->image_view);
-  vk_check_error ("vkCreateImageView", res, NULL)
+  vk_check_error ("vkCreateImageView", res, NULL);
 
-  if (!gulkan_device_create_buffer (device,
-                                    size,
-                                    VK_BUFFER_USAGE_TRANSFER_SRC_BIT,
-                                    VK_MEMORY_PROPERTY_HOST_VISIBLE_BIT,
-                                   &self->staging_buffer,
-                                   &self->staging_buffer_memory))
-    return NULL;
   return self;
 }
 
-static bool
-_upload_pixels (GulkanTexture           *self,
-                VkCommandBuffer          cmd_buffer,
-                guchar                  *pixels,
-                gsize                    size,
-                const VkBufferImageCopy *regions,
-                VkImageLayout            layout)
-{
-  if (!gulkan_device_map_memory (self->device, pixels, size,
-                                 self->staging_buffer_memory))
-    return false;
-
-  gulkan_texture_transfer_layout (self, self->device, cmd_buffer,
-                                  VK_IMAGE_LAYOUT_UNDEFINED,
-                                  VK_IMAGE_LAYOUT_TRANSFER_DST_OPTIMAL);
-
-  vkCmdCopyBufferToImage (cmd_buffer,
-                          self->staging_buffer, self->image,
-                          VK_IMAGE_LAYOUT_TRANSFER_DST_OPTIMAL,
-                          self->mip_levels,
-                          regions);
-
-  gulkan_texture_transfer_layout (self, self->device, cmd_buffer,
-                                  VK_IMAGE_LAYOUT_TRANSFER_DST_OPTIMAL,
-                                  layout);
-
-  return true;
-}
-
 static GulkanMipMap
 _generate_mipmaps (GdkPixbuf *pixbuf)
 {
@@ -370,7 +371,7 @@ _generate_mipmaps (GdkPixbuf *pixbuf)
       mip_height /= 2;
 
       GdkPixbuf *level_pixbuf = gdk_pixbuf_new (GDK_COLORSPACE_RGB,
-                                               true, 8,
+                                               TRUE, 8,
                                                MAX (mip_width, 1),
                                                MAX (mip_height, 1));
 
@@ -407,9 +408,8 @@ _generate_mipmaps (GdkPixbuf *pixbuf)
   return mipmap;
 }
 
-bool
+gboolean
 gulkan_texture_upload_pixels (GulkanTexture  *self,
-                              VkCommandBuffer cmd_buffer,
                               guchar         *pixels,
                               gsize           size,
                               VkImageLayout   layout)
@@ -417,7 +417,7 @@ gulkan_texture_upload_pixels (GulkanTexture  *self,
   if (self->mip_levels != 1)
     {
       g_warning ("Trying to upload one mip level to multi level texture.\n");
-      return false;
+      return FALSE;
     }
 
   VkBufferImageCopy buffer_image_copy = {
@@ -428,29 +428,50 @@ gulkan_texture_upload_pixels (GulkanTexture  *self,
       .aspectMask = VK_IMAGE_ASPECT_COLOR_BIT,
     },
     .imageExtent = {
-      .width = self->width,
-      .height = self->height,
+      .width = self->extent.width,
+      .height = self->extent.height,
       .depth = 1,
     }
   };
 
-  return _upload_pixels (self, cmd_buffer, pixels, size,
-                        &buffer_image_copy, layout);
+  return _upload_pixels (self, pixels, size, &buffer_image_copy, layout);
 }
 
+gboolean
+gulkan_texture_upload_pixbuf (GulkanTexture *self,
+                              GdkPixbuf     *pixbuf,
+                              VkImageLayout  layout)
+{
+  guchar *pixels = gdk_pixbuf_get_pixels (pixbuf);
+  gsize size = gdk_pixbuf_get_byte_length (pixbuf);
+
+  return gulkan_texture_upload_pixels (self, pixels, size, layout);
+}
+
+gboolean
+gulkan_texture_upload_cairo_surface (GulkanTexture   *self,
+                                     cairo_surface_t *surface,
+                                     VkImageLayout    layout)
+{
+  guchar *pixels = cairo_image_surface_get_data (surface);
+  gsize size = (gsize) cairo_image_surface_get_stride (surface) *
+               (gsize) cairo_image_surface_get_height (surface);
+
+  return gulkan_texture_upload_pixels (self, pixels, size, layout);
+}
+
+
 GulkanTexture *
-gulkan_texture_new_from_dmabuf (GulkanDevice *device,
+gulkan_texture_new_from_dmabuf (GulkanClient *client,
                                 int           fd,
-                                guint         width,
-                                guint         height,
+                                VkExtent2D    extent,
                                 VkFormat      format)
 {
   GulkanTexture *self = (GulkanTexture*) g_object_new (GULKAN_TYPE_TEXTURE, 0);
-  VkDevice vk_device = gulkan_device_get_handle (device);
+  VkDevice vk_device = gulkan_client_get_device_handle (client);
 
-  self->width = width;
-  self->height = height;
-  self->device = device;
+  self->extent = extent;
+  self->client = g_object_ref (client);
   self->format = format;
 
   VkExternalMemoryImageCreateInfoKHR external_memory_image_create_info = {
@@ -463,8 +484,8 @@ gulkan_texture_new_from_dmabuf (GulkanDevice *device,
     .pNext = &external_memory_image_create_info,
     .imageType = VK_IMAGE_TYPE_2D,
     .extent = {
-      .width = width,
-      .height = height,
+      .width = extent.width,
+      .height = extent.height,
       .depth = 1,
     },
     .mipLevels = 1,
@@ -481,7 +502,7 @@ gulkan_texture_new_from_dmabuf (GulkanDevice *device,
 
   VkResult res;
   res = vkCreateImage (vk_device, &image_info, NULL, &self->image);
-  vk_check_error ("vkCreateImage", res, NULL)
+  vk_check_error ("vkCreateImage", res, NULL);
 
   VkMemoryDedicatedAllocateInfoKHR dedicated_memory_info = {
     .sType = VK_STRUCTURE_TYPE_MEMORY_DEDICATED_ALLOCATE_INFO_KHR,
@@ -497,7 +518,7 @@ gulkan_texture_new_from_dmabuf (GulkanDevice *device,
     .fd = fd
   };
 
-  VkMemoryRequirements memory_requirements = {};
+  VkMemoryRequirements memory_requirements;
   vkGetImageMemoryRequirements (vk_device, self->image,
                                 &memory_requirements);
 
@@ -509,17 +530,17 @@ gulkan_texture_new_from_dmabuf (GulkanDevice *device,
   };
 
   gulkan_device_memory_type_from_properties (
-    device,
+    gulkan_client_get_device (client),
     memory_requirements.memoryTypeBits,
     VK_MEMORY_PROPERTY_HOST_CACHED_BIT,
     &memory_info.memoryTypeIndex);
 
   res = vkAllocateMemory (vk_device, &memory_info,
                           NULL, &self->image_memory);
-  vk_check_error ("vkAllocateMemory", res, NULL)
+  vk_check_error ("vkAllocateMemory", res, NULL);
 
   res = vkBindImageMemory (vk_device, self->image, self->image_memory, 0);
-  vk_check_error ("vkBindImageMemory", res, NULL)
+  vk_check_error ("vkBindImageMemory", res, NULL);
 
   VkImageViewCreateInfo image_view_info =
   {
@@ -538,17 +559,17 @@ gulkan_texture_new_from_dmabuf (GulkanDevice *device,
   };
   res = vkCreateImageView (vk_device, &image_view_info,
                            NULL, &self->image_view);
-  vk_check_error ("vkCreateImageView", res, NULL)
+  vk_check_error ("vkCreateImageView", res, NULL);
 
   return self;
 }
 
 /**
  * gulkan_texture_new_export_fd:
- * @device: a #GulkanDevice
- * @width: Width in pixels
- * @height: Height in pixels
+ * @client: a #GulkanClient
+ * @extent: Extent in pixels
  * @format: VkFormat of the texture
+ * @layout: VkImageLayout of the texture
  * @size: Return value of allocated size
  * @fd: Return value for allocated fd
  *
@@ -562,19 +583,18 @@ gulkan_texture_new_from_dmabuf (GulkanDevice *device,
  * Returns: the initialized #GulkanTexture
  */
 GulkanTexture *
-gulkan_texture_new_export_fd (GulkanDevice *device,
-                              guint         width,
-                              guint         height,
+gulkan_texture_new_export_fd (GulkanClient *client,
+                              VkExtent2D    extent,
                               VkFormat      format,
+                              VkImageLayout layout,
                               gsize        *size,
                               int          *fd)
 {
   GulkanTexture *self = (GulkanTexture*) g_object_new (GULKAN_TYPE_TEXTURE, 0);
-  VkDevice vk_device = gulkan_device_get_handle (device);
+  VkDevice vk_device = gulkan_client_get_device_handle (client);
 
-  self->width = width;
-  self->height = height;
-  self->device = device;
+  self->extent = extent;
+  self->client = g_object_ref (client);
   self->format = format;
 
   /* we can also export the memory of the image without using this struct but
@@ -590,8 +610,8 @@ gulkan_texture_new_export_fd (GulkanDevice *device,
     .pNext = &external_memory_image_create_info,
     .imageType = VK_IMAGE_TYPE_2D,
     .extent = {
-      .width = width,
-      .height = height,
+      .width = extent.width,
+      .height = extent.height,
       .depth = 1,
     },
     .mipLevels = 1,
@@ -600,14 +620,15 @@ gulkan_texture_new_export_fd (GulkanDevice *device,
     .tiling = VK_IMAGE_TILING_OPTIMAL,
     .samples = VK_SAMPLE_COUNT_1_BIT,
     .usage = VK_IMAGE_USAGE_SAMPLED_BIT |
-             VK_IMAGE_USAGE_TRANSFER_SRC_BIT,
+             VK_IMAGE_USAGE_TRANSFER_SRC_BIT |
+             VK_IMAGE_USAGE_TRANSFER_DST_BIT,
     .sharingMode = VK_SHARING_MODE_EXCLUSIVE,
     .initialLayout = VK_IMAGE_LAYOUT_UNDEFINED
   };
 
   VkResult res;
   res = vkCreateImage (vk_device, &image_info, NULL, &self->image);
-  vk_check_error ("vkCreateImage", res, NULL)
+  vk_check_error ("vkCreateImage", res, NULL);
 
   VkMemoryDedicatedAllocateInfoKHR dedicated_memory_info =
   {
@@ -617,7 +638,7 @@ gulkan_texture_new_export_fd (GulkanDevice *device,
     .buffer = VK_NULL_HANDLE
   };
 
-  VkMemoryRequirements memory_reqs = {};
+  VkMemoryRequirements memory_reqs;
   vkGetImageMemoryRequirements (vk_device, self->image,
                                 &memory_reqs);
 
@@ -636,7 +657,9 @@ gulkan_texture_new_export_fd (GulkanDevice *device,
     VK_MEMORY_PROPERTY_DEVICE_LOCAL_BIT
     | VK_MEMORY_PROPERTY_HOST_COHERENT_BIT;
 
-  bool full_flags_available =
+  GulkanDevice *device = gulkan_client_get_device (client);
+
+  gboolean full_flags_available =
     gulkan_device_memory_type_from_properties (device,
                                                memory_reqs.memoryTypeBits,
                                                full_memory_property_flags,
@@ -660,10 +683,10 @@ gulkan_texture_new_export_fd (GulkanDevice *device,
 
   res = vkAllocateMemory (vk_device, &memory_info,
                           NULL, &self->image_memory);
-  vk_check_error ("vkAllocateMemory", res, NULL)
+  vk_check_error ("vkAllocateMemory", res, NULL);
 
   res = vkBindImageMemory (vk_device, self->image, self->image_memory, 0);
-  vk_check_error ("vkBindImageMemory", res, NULL)
+  vk_check_error ("vkBindImageMemory", res, NULL);
 
 
   VkImageViewCreateInfo image_view_info =
@@ -683,17 +706,40 @@ gulkan_texture_new_export_fd (GulkanDevice *device,
   };
   res = vkCreateImageView (vk_device, &image_view_info,
                            NULL, &self->image_view);
-  vk_check_error ("vkCreateImageView", res, NULL)
+  vk_check_error ("vkCreateImageView", res, NULL);
 
-  if (!gulkan_device_gulkan_device_get_memory_fd (device,
-                                                  self->image_memory,
-                                                  fd))
+  if (!gulkan_device_get_memory_fd (device, self->image_memory, fd))
     {
       g_printerr ("Could not get file descriptor for memory!\n");
       g_object_unref (self);
       return NULL;
     }
-  /* g_print ("got FD for texture %d\n", *fd); */
+
+  GulkanQueue *queue = gulkan_device_get_transfer_queue (device);
+  GulkanCmdBuffer *cmd_buffer = gulkan_queue_request_cmd_buffer (queue);
+  GMutex *mutex = gulkan_queue_get_pool_mutex (queue);
+
+  g_mutex_lock (mutex);
+  gboolean ret = gulkan_cmd_buffer_begin (cmd_buffer);
+  g_mutex_unlock (mutex);
+  if (!ret)
+    return FALSE;
+
+  gulkan_texture_record_transfer (self,
+                                  gulkan_cmd_buffer_get_handle (cmd_buffer),
+                                  VK_IMAGE_LAYOUT_UNDEFINED,
+                                  VK_IMAGE_LAYOUT_TRANSFER_DST_OPTIMAL);
+
+  gulkan_texture_record_transfer (self,
+                                  gulkan_cmd_buffer_get_handle (cmd_buffer),
+                                  VK_IMAGE_LAYOUT_TRANSFER_DST_OPTIMAL,
+                                  layout);
+
+  if (!gulkan_queue_submit (queue, cmd_buffer))
+    return FALSE;
+
+  gulkan_queue_free_cmd_buffer (queue, cmd_buffer);
+
   return self;
 }
 
@@ -723,13 +769,13 @@ _get_access_flags (VkImageLayout layout)
 }
 
 void
-gulkan_texture_transfer_layout (GulkanTexture       *self,
-                                GulkanDevice        *device,
+gulkan_texture_record_transfer (GulkanTexture       *self,
                                 VkCommandBuffer      cmd_buffer,
                                 VkImageLayout        src_layout,
                                 VkImageLayout        dst_layout)
 {
-  gulkan_texture_transfer_layout_full (self, device, cmd_buffer,
+  gulkan_texture_record_transfer_full (self,
+                                       cmd_buffer,
                                        _get_access_flags (src_layout),
                                        _get_access_flags (dst_layout),
                                        src_layout, dst_layout,
@@ -738,8 +784,7 @@ gulkan_texture_transfer_layout (GulkanTexture       *self,
 }
 
 void
-gulkan_texture_transfer_layout_full (GulkanTexture       *self,
-                                     GulkanDevice        *device,
+gulkan_texture_record_transfer_full (GulkanTexture       *self,
                                      VkCommandBuffer      cmd_buffer,
                                      VkAccessFlags        src_access_mask,
                                      VkAccessFlags        dst_access_mask,
@@ -748,6 +793,11 @@ gulkan_texture_transfer_layout_full (GulkanTexture       *self,
                                      VkPipelineStageFlags src_stage_mask,
                                      VkPipelineStageFlags dst_stage_mask)
 {
+  GulkanDevice *device = gulkan_client_get_device (self->client);
+  GulkanQueue *gulkan_queue = gulkan_device_get_transfer_queue (device);
+  uint32_t queue_index = gulkan_queue_get_family_index (gulkan_queue);
+  GMutex *mutex = gulkan_queue_get_pool_mutex (gulkan_queue);
+
   VkImageMemoryBarrier image_memory_barrier =
   {
     .sType = VK_STRUCTURE_TYPE_IMAGE_MEMORY_BARRIER,
@@ -763,15 +813,78 @@ gulkan_texture_transfer_layout_full (GulkanTexture       *self,
       .baseArrayLayer = 0,
       .layerCount = 1,
     },
-    .srcQueueFamilyIndex = gulkan_device_get_queue_family_index (device),
-    .dstQueueFamilyIndex = gulkan_device_get_queue_family_index (device)
+    .srcQueueFamilyIndex = queue_index,
+    .dstQueueFamilyIndex = queue_index
   };
 
+  g_mutex_lock (mutex);
   vkCmdPipelineBarrier (cmd_buffer,
                         src_stage_mask,
                         dst_stage_mask,
                         0, 0, NULL, 0, NULL, 1,
                         &image_memory_barrier);
+  g_mutex_unlock (mutex);
+}
+
+gboolean
+gulkan_texture_transfer_layout_full (GulkanTexture       *self,
+                                     VkAccessFlags        src_access_mask,
+                                     VkAccessFlags        dst_access_mask,
+                                     VkImageLayout        src_layout,
+                                     VkImageLayout        dst_layout,
+                                     VkPipelineStageFlags src_stage_mask,
+                                     VkPipelineStageFlags dst_stage_mask)
+{
+  GulkanDevice *device = gulkan_client_get_device (self->client);
+  GulkanQueue *queue = gulkan_device_get_transfer_queue (device);
+  GulkanCmdBuffer *cmd_buffer = gulkan_queue_request_cmd_buffer (queue);
+  GMutex *mutex = gulkan_queue_get_pool_mutex (queue);
+  g_mutex_lock (mutex);
+  gboolean ret = gulkan_cmd_buffer_begin (cmd_buffer);
+  g_mutex_unlock (mutex);
+  if (!ret)
+    return FALSE;
+
+  gulkan_texture_record_transfer_full (self,
+                                       gulkan_cmd_buffer_get_handle (cmd_buffer),
+                                       src_access_mask, dst_access_mask,
+                                       src_layout, dst_layout,
+                                       src_stage_mask, dst_stage_mask);
+
+  if (!gulkan_queue_submit (queue, cmd_buffer))
+    return FALSE;
+
+  gulkan_queue_free_cmd_buffer (queue, cmd_buffer);
+
+  return TRUE;
+}
+
+gboolean
+gulkan_texture_transfer_layout (GulkanTexture *self,
+                                VkImageLayout  src_layout,
+                                VkImageLayout  dst_layout)
+{
+  GulkanDevice *device = gulkan_client_get_device (self->client);
+  GulkanQueue *queue = gulkan_device_get_transfer_queue (device);
+  GulkanCmdBuffer *cmd_buffer = gulkan_queue_request_cmd_buffer (queue);
+  GMutex *mutex = gulkan_queue_get_pool_mutex (queue);
+  g_mutex_lock (mutex);
+  gboolean ret = gulkan_cmd_buffer_begin (cmd_buffer);
+  g_mutex_unlock (mutex);
+
+  if (!ret)
+    return FALSE;
+
+  gulkan_texture_record_transfer (self,
+                                  gulkan_cmd_buffer_get_handle (cmd_buffer),
+                                  src_layout, dst_layout);
+
+  if (!gulkan_queue_submit (queue, cmd_buffer))
+    return FALSE;
+
+  gulkan_queue_free_cmd_buffer (queue, cmd_buffer);
+
+  return TRUE;
 }
 
 VkImageView
@@ -786,16 +899,10 @@ gulkan_texture_get_image (GulkanTexture *self)
   return self->image;
 }
 
-guint
-gulkan_texture_get_width (GulkanTexture *self)
+VkExtent2D
+gulkan_texture_get_extent (GulkanTexture *self)
 {
-  return self->width;
-}
-
-guint
-gulkan_texture_get_height (GulkanTexture *self)
-{
-  return self->height;
+  return self->extent;
 }
 
 VkFormat
