@@ -16,7 +16,7 @@
 
 #include "gulkan-version.h"
 #include "gulkan-client.h"
-#include "gulkan-descriptor-set.h"
+#include "gulkan-cmd-buffer.h"
 #include "gulkan-device.h"
 #include "gulkan-frame-buffer.h"
 #include "gulkan-geometry.h"
@@ -25,6 +25,12 @@
 #include "gulkan-texture.h"
 #include "gulkan-uniform-buffer.h"
 #include "gulkan-vertex-buffer.h"
+#include "gulkan-render-pass.h"
+#include "gulkan-swapchain.h"
+#include "gulkan-descriptor-pool.h"
+#include "gulkan-swapchain-renderer.h"
+#include "gulkan-buffer.h"
+#include "gulkan-queue.h"
 
 #undef GULKAN_INSIDE
 
