@@ -16,7 +16,7 @@ load_gdk_pixbuf ()
 
   g_assert_null (error);
 
-  GdkPixbuf *pixbuf = gdk_pixbuf_add_alpha (pixbuf_rgb, false, 0, 0, 0);
+  GdkPixbuf *pixbuf = gdk_pixbuf_add_alpha (pixbuf_rgb, FALSE, 0, 0, 0);
   g_object_unref (pixbuf_rgb);
   return pixbuf;
 }
@@ -38,13 +38,11 @@ _test_pixbuf_texture ()
   GulkanClient *client = gulkan_client_new ();
   g_assert_nonnull (client);
 
-  g_assert (gulkan_client_init_vulkan (client, NULL, NULL));
-
   GulkanTexture *texture =
-    gulkan_client_texture_new_from_pixbuf (client, pixbuf,
-                                           VK_FORMAT_R8G8B8A8_UNORM,
-                                           VK_IMAGE_LAYOUT_TRANSFER_SRC_OPTIMAL,
-                                           false);
+    gulkan_texture_new_from_pixbuf (client, pixbuf,
+                                    VK_FORMAT_R8G8B8A8_UNORM,
+                                    VK_IMAGE_LAYOUT_TRANSFER_SRC_OPTIMAL,
+                                    FALSE);
   g_assert_nonnull (texture);
 
   g_object_unref (texture);
